"""
This module contains the constants used in the application.
"""

INFO_METRIC_NAME = "info"
CONNECTION_METRIC_NAME = "connection_up"
READ_LATENCY_METRIC_NAME = "read_time_seconds"
ERROR_FREQUENCY_METRIC_NAME = "error_count"
DEFAULT_SCRAPE_INTERVAL = "30s"
DEFAULT_PORT = 9075
DEFAULT_LOG_LEVEL = "info"
DEFAULT_NAMESPACE = "plc"
DEFAULT_IDENTIFER = "master"
DEFAULT_METRIC_LAYOUT = "dynamic"
